# The naming conventions follow those used in github.com/AstroDigitalVideo/ADVlib
# While not exactly 'Pythonic' (i.e., not snake case), it was my judgement (Bob Anderson) that
# fewer errors would be introduced and would reduce the intellectual 'load' of comparing
# the Python code against the equivalent code in github.com/AstroDigitalVideo/ADVlib

# We use type hinting so that it is easy to see the intent as matching the C++ code

import os

# import sys
# print('\nsys.path ...')
# print('\n'.join(sys.path))
# print('\ncurrent working directory ...')
# print(os.getcwd(), '\n')

import pathlib
from ctypes import c_int, c_uint
from typing import Dict, List, Tuple
from datetime import datetime, timedelta
import numpy as np
from Adv2.AdvError import ResolveErrorMessage, S_OK
from Adv2 import AdvLib
from Adv2.Adv import AdvFileInfo, AdvFrameInfo, AdvIndexEntry, StreamId, TagPairType, Adv2TagType
from Adv2.AdvError import AdvLibException


class Adv2reader:
    def __init__(self, filename: str):
        if not os.path.isfile(filename):
            raise AdvLibException(f'Error: cannot find file ... {filename}')

        # Get version number without trying to fully parse the file.
        err, version = AdvLib.AdvGetFileVersion(filename)
        if not err:
            if not version == 2:
                raise AdvLibException(f'ADV version {version} is unsupported --- only version 2 is supported')
        else:
            raise AdvLibException(err)

        fileInfo = AdvFileInfo()
        fileVersionErrorCode = AdvLib.AdvOpenFile(filename, fileInfo)

        if fileVersionErrorCode > 0x70000000:
            raise AdvLibException(f'There was an error opening {filename}: '
                                  f'{ResolveErrorMessage(fileVersionErrorCode)}')

        if not fileVersionErrorCode == 2:
            raise AdvLibException(f'{filename} is not an ADV version 2 file')

        if fileInfo.IsColourImage:
            raise AdvLibException(f'This file contains color images: not yet supported.')

        # Provide a few useful top-level instance variables
        self.Width = fileInfo.Width
        self.Height = fileInfo.Height
        self.CountMainFrames = fileInfo.CountMainFrames
        self.CountCalibrationFrames = fileInfo.CountCalibrationFrames

        # This instance variable gives the user access all file information, including the above items
        self.FileInfo = fileInfo

        self.pixels = None
        self.sysErrLength: int = 0
        self.frameInfo = AdvFrameInfo()
        self.statusTagInfo = []

        for tagId in range(fileInfo.StatusTagsCount):
            tagType, tagName = AdvLib.AdvVer2_GetStatusTagInfo(tagId)
            if tagName:
                self.statusTagInfo.append((tagType, tagName))

    def getMainImageAndStatusData(self, frameNumber: int) -> Tuple[str, np.ndarray, AdvFrameInfo, Dict[str, any]]:
        return self._getGenericImageAndStatusData(frameNumber=frameNumber, streamType=StreamId.Main)

    def getCalibImageAndStatusData(self, frameNumber: int) -> Tuple[str, np.ndarray, AdvFrameInfo, Dict[str, any]]:
        return self._getGenericImageAndStatusData(frameNumber=frameNumber, streamType=StreamId.Calibration)

    def _getGenericImageAndStatusData(self, frameNumber: int, streamType: StreamId) -> \
            Tuple[str, np.ndarray, AdvFrameInfo, Dict[str, any]]:
        err_msg = ''
        # Create an array of c_uint to hold the pixel values
        pixel_array = (c_uint * (self.Width * self.Height))()

        ret_val = AdvLib.AdvVer2_GetFramePixels(
            streamId=streamType, frameNo=frameNumber,
            pixels=pixel_array, frameInfo=self.frameInfo, systemErrorLen=self.sysErrLength
        )

        # Convert pixel_array to 1D numpy array, then reshape into 2D numpy array
        self.pixels = np.reshape(np.array(pixel_array, dtype='uint16'), newshape=(self.Height, self.Width))

        if ret_val is not S_OK:
            err_msg = ResolveErrorMessage(ret_val)
            return err_msg, self.pixels, self.frameInfo, {}

        # This code block adds date and start-of-exposure timestamp strings to self.frameInfo

        datetime64 = self.frameInfo.UtcMidExposureTimestampLo + (self.frameInfo.UtcMidExposureTimestampHi << 32)
        # Adjust from mid exposure to start-exposure
        datetime64 -= self.frameInfo.Exposure // 2

        usecs = datetime64 // 1000  # Convert nanoseconds to microseconds - needed for call to timedelta
        ts = datetime(2010, 1, 1) + timedelta(microseconds=usecs)

        # There 'should' always be a valid timestamp in an adv2 file, but if there is not, then there will be
        # zeros in UtcMidExposureTimestampLo and Hi.  We return empty strings in this case.
        if timedelta == 0:
            self.frameInfo.DateString = ''
            self.frameInfo.StartOfExposureTimestampString = ''
        else:
            self.frameInfo.DateString = f'{ts.year:04d}-{ts.month:02d}-{ts.day:02d}'
            # This string below is in the form needed for direct insertion into a csv file column (Excel safe format)
            self.frameInfo.StartOfExposureTimestampString = \
                f'[{ts.hour:02d}:{ts.minute:02d}:{ts.second:02d}.{ts.microsecond:06d}]'

        # end code block

        status_dict = {}
        for i in range(len(self.statusTagInfo)):
            tagType, tagName = self.statusTagInfo[i]
            if tagType == Adv2TagType.Int8:
                val = AdvLib.AdvVer2_GetStatusTagUInt8(i)
                status_dict.update({tagName: val})
            if tagType == Adv2TagType.Int16:
                val = AdvLib.AdvVer2_GetStatusTagInt16(i)
                status_dict.update({tagName: val})
            if tagType == Adv2TagType.Int32:
                val = AdvLib.AdvVer2_GetStatusTagInt32(i)
                status_dict.update({tagName: val})
            if tagType == Adv2TagType.Int64:
                val = AdvLib.AdvVer2_GetStatusTagInt64(i)
                if tagName == 'SystemTime':
                    usecs = val // 1000  # Convert nanoseconds to microseconds - for timedelta
                    ts = datetime(2010, 1, 1) + timedelta(microseconds=usecs)
                    val_str = (f'{ts.year:04d}-{ts.month:02d}-{ts.day:02d} '
                               f'[{ts.hour:02d}:{ts.minute:02d}:{ts.second:02d}.{ts.microsecond:06d}]')
                    status_dict.update({tagName: val_str})
                else:
                    status_dict.update({tagName: val})
            if tagType == Adv2TagType.Real:
                val = AdvLib.AdvVer2_GetStatusTagReal(i)
                status_dict.update({tagName: val})
            if tagType == Adv2TagType.UTF8String:
                val = AdvLib.AdvVer2_GetStatusTagUTF8String(i)
                status_dict.update({tagName: val})

        return err_msg, self.pixels, self.frameInfo, status_dict

    def getAdvFileMetaData(self) -> Dict[str, str]:
        meta_dict = {}
        if self.FileInfo.SystemMetadataTagsCount > 0:
            for entryNum in range(self.FileInfo.SystemMetadataTagsCount):
                err_msg, name, value = AdvLib.AdvVer2_GetTagPairValues(TagPairType.SystemMetaData, entryNum)
                if not err_msg:
                    meta_dict.update({name: value})
        if self.FileInfo.UserMetadataTagsCount > 0:
            for entryNum in range(self.FileInfo.UserMetadataTagsCount):
                err_msg, name, value = AdvLib.AdvVer2_GetTagPairValues(TagPairType.UserMetaData, entryNum)
                if not err_msg:
                    meta_dict.update({name: value})
        return meta_dict

    def getIndexEntries(self) -> Tuple[List[AdvIndexEntry], List[AdvIndexEntry]]:
        # Create C compatible buffers that hold the correct number of AdvIndexEntry instances.
        # Although an AdvIndexEntry is 2 int64 and an int32, we have to allow for 'alignment' on int64
        # boundaries - that's the reason for the 6 c_int rather than 5 c_int in the buffer construction
        mainIndex = (c_int * (6 * self.CountMainFrames))()
        calibIndex = (c_int * (6 * self.CountCalibrationFrames))()

        ret_val = AdvLib.AdvVer2_GetIndexEntries(mainIndex, calibIndex)

        if ret_val is S_OK:
            mainList = []
            calibList = []

            base = 0
            for _ in range(self.CountMainFrames):
                ticks = mainIndex[base] + (mainIndex[base + 1] << 32)
                offset = mainIndex[base + 2] + (mainIndex[base + 3] << 32)
                byte_count = mainIndex[base + 4]

                new_index_entry = AdvIndexEntry()
                new_index_entry.ElapsedTicks = ticks
                new_index_entry.FrameOffset = offset
                new_index_entry.BytesCount = byte_count

                mainList.append(new_index_entry)
                base += 6

            base = 0
            for _ in range(self.CountCalibrationFrames):
                ticks = calibIndex[base] + (calibIndex[base + 1] << 32)
                offset = calibIndex[base + 2] + (calibIndex[base + 3] << 32)
                byte_count = calibIndex[base + 4]

                new_index_entry = AdvIndexEntry()
                new_index_entry.ElapsedTicks = ticks
                new_index_entry.FrameOffset = offset
                new_index_entry.BytesCount = byte_count

                calibList.append(new_index_entry)
                base += 6

            return mainList, calibList
        else:
            raise AdvLibException(f'{ResolveErrorMessage(ret_val)}')

    @staticmethod
    def closeFile():
        return AdvLib.AdvCloseFile()


def exerciser():
    import sys
    import cv2  # Used by exerciser() only

    default_file = pathlib.Path(__file__).parent / 'UnitTestSample.adv'
    # default_file = pathlib.Path(__file__).parent / '2012-01-02 01h03m(0036).adv'  # Version 1 test file

    num_frames_to_view = 6  # This currently the number of frames in the UnitTestSample.adv
    file_to_use = default_file
    if len(sys.argv) == 1:
        print('\nNo command line arguments supplied. Defaults will be used.')
    elif len(sys.argv) >= 2:
        file_to_use = pathlib.Path(sys.argv[1])
        if len(sys.argv) > 2:
            num_frames_to_view = int(sys.argv[2])

    print(f'\nfile to be viewed: {file_to_use}')
    print(f'{num_frames_to_view} frames are to be viewed\n')

    print(f'\nGeneral information:')
    print(f'    library: {AdvLib.GetLibraryVersion()}')
    print(f'    platform: {AdvLib.GetLibraryPlatformId()}')
    print(f'    bitness: {AdvLib.GetLibraryBitness()}\n')

    # Get version number without trying to fully parse the file (which instantiating Adv2reader() does)
    err, version = AdvLib.AdvGetFileVersion(str(file_to_use))
    if not err:
        if not version == 2:
            print(f'ADV version {version} is unsupported --- only version 2 is supported')
            return
        else:
            print(f'ADV file version {version} found\n')
    else:
        print(err)
        return

    rdr = None
    try:
        file_path = str(file_to_use)
        rdr = Adv2reader(file_path)
    except AdvLibException as adverr:
        print(repr(adverr))
        exit()

    # Show some top level instance variables
    print(f'A few top-level instance variables of general interest:')
    print(f'    Width: {rdr.Width}  Height: {rdr.Height}  NumMainFrames: {rdr.CountMainFrames}'
          f'  NumCalibFrames: {rdr.CountCalibrationFrames}')
    print(f'    Color image: {rdr.FileInfo.IsColourImage}')
    mainIndexList, calibIndexList = rdr.getIndexEntries()
    print(f'    mainIndexList has {len(mainIndexList)} entries')
    print(f'    calibIndexList has {len(calibIndexList)} entries\n')

    # Show the status tag name and types associated with each DATA_FRAME This is the layout of the STATUS_SECTION
    print(f'STATUS_SECTION layout:')
    for entry in rdr.statusTagInfo:
        tagType, tagName = entry
        print(f'    {tagName}: {tagType}')

    print(f'\nADV_FILE_META_DATA:')
    meta_data = rdr.getAdvFileMetaData()
    for key in meta_data:
        print(f'    {key}: {meta_data[key]}')

    # Show a few main index entries
    # for i in range(len(mainIndexList)):
    # for i in range(3):
    #     print(f'\nindex: {i:2d} ElapsedTicks: {mainIndexList[i].ElapsedTicks}')
    #     print(f'index: {i:2d}  FrameOffset: {mainIndexList[i].FrameOffset}')
    #     print(f'index: {i:2d}   BytesCount: {mainIndexList[i].BytesCount}')

    # For display purposes, we need to scale data that BITPIX less than 16 to a full 16 bit scale
    scale_factor = 1 << (16 - int(meta_data.setdefault('BITPIX', '16')))
    if scale_factor > 1:
        print(f'\nThe image data has been scaled (multiplied) by {scale_factor} because bits/pixel is < 16')

    if rdr.Width < 300:
        zoom_factor = 20
        print(f'A zoom factor of {zoom_factor} has been applied to the image to aid visibility.\n')

    print(f'\n!!!!  Press any key to advance to next frame (with image selected)  !!!!')

    # for frame in range(rdr.CountMainFrames):
    for frame in range(num_frames_to_view):
        err, image, frameInfo, status = rdr.getMainImageAndStatusData(frameNumber=frame)

        image = image * scale_factor

        # If we're running a UnitTestSample with tiny images (too conserve space), we automatically zoom the image
        if rdr.Width < 300:
            zoom_factor = 20
            image = np.repeat(image, zoom_factor, axis=0)
            image = np.repeat(image, zoom_factor, axis=1)

        if not err:
            print(f'\nframe: {frame} STATUS:')
            # print(frameInfo.DateString, frameInfo.StartOfExposureTimestampString)
            # print(f'RawDataSize: {frameInfo.RawDataBlockSize}')
            for k in status:
                print(f'    {k}: {status[k]}')
            cv2.imshow('Test image', image)
            cv2.waitKey(0)  # You must press a key while the image is topmost to advance
        else:
            print(err)

    # print(f'\nimage.shape: {image.shape}  image.dtype: {image.dtype}\n')
    print(f'\ncloseFile returned: {rdr.closeFile()}')
    # print(f'\nun-needed closeFile returned: {rdr.closeFile()}\n')


if __name__ == "__main__":
    exerciser()



# The naming conventions follow those used in github.com/AstroDigitalVideo/ADVlib
# While not exactly 'Pythonic' (i.e., not snake case), it was my judgement (Bob Anderson) that
# fewer errors would be introduced and would reduce the intellectual 'load' of comparing
# the Python code against the equivalent code in github.com/AstroDigitalVideo/ADVlib

# This module holds all of the enums and structures that are defined
# in the C# file PInvoke.cs in the namespace Adv (equivalent to a Python module/file)

# We use type hinting so that it is easy to see the intent as matching the C++/C# code

from dataclasses import dataclass
from enum import Enum


class StreamId(Enum):
    Main = 0
    Calibration = 1


class Adv2TagType(Enum):
    Int8 = 0
    Int16 = 1
    Int32 = 2
    Int64 = 3
    Real = 4
    UTF8String = 5


class TagPairType(Enum):
    MainStream = 0
    CalibrationStream = 1
    SystemMetaData = 2
    UserMetaData = 3
    ImageSection = 4
    FirstImageLayout = 100


@dataclass
class AdvImageLayoutInfo:
    ImageLayoutId: int = 0
    ImageLayoutTagsCount: int = 0
    ImageLayoutBpp: int = 0
    IsFullImageRaw: bool = False
    Is12BitImagePacked: bool = False
    Is8BitColourPacked: bool = False


@dataclass
class AdvFileInfo:
    Width: int = 0
    Height: int = 0
    CountMainFrames: int = 0  # spelling change here --- is CountMaintFrames in C# code
    CountCalibrationFrames: int = 0
    DataBpp: int = 0
    MaxPixelValue: int = 0
    MainClockFrequency: int = 0
    MainStreamAccuracy: int = 0
    CalibrationClockFrequency: int = 0
    CalibrationStreamAccuracy: int = 0
    MainStreamTagsCount: int = 0
    CalibrationStreamTagsCount: int = 0
    SystemMetadataTagsCount: int = 0
    UserMetadataTagsCount: int = 0
    UtcTimestampAccuracyInNanoseconds: int = 0
    IsColourImage: bool = False
    ImageLayoutsCount: int = 0
    StatusTagsCount: int = 0
    ImageSectionTagsCount: int = 0
    ErrorStatusTagId: int = 0


@dataclass
class AdvIndexEntry:
    ElapsedTicks: int = 0
    FrameOffset: int = 0
    BytesCount: int = 0


@dataclass
class AdvFrameInfo:
    StartTicksLo: int = 0
    StartTicksHi: int = 0
    EndTicksLo: int = 0
    EndTicksHi: int = 0

    UtcMidExposureTimestampLo: int = 0
    UtcMidExposureTimestampHi: int = 0
    Exposure: int = 0

    DateString: str = ''                   # This a new item, not included in C# AdvFrameInfo
    StartOfExposureTimestampString = ''    # This a new item, not included in C# AdvFrameInfo

    Gamma: float = 0.0
    Gain: float = 0.0
    Shutter: float = 0.0
    Offset: float = 0.0

    GPSTrackedSatellites: int = 0  # Spelling correction for satellites applied here
    GPSAlmanacStatus: int = 0
    GPSFixStatus: int = 0
    GPSAlmanacOffset: int = 0

    VideoCameraFrameIdLo: int = 0
    VideoCameraFrameIdHi: int = 0
    HardwareTimerFrameIdLo: int = 0
    HardwareTimerFrameIdHi = 0

    SystemTimestampLo: int = 0
    SystemTimestampHi: int = 0

    ImageLayoutId = 0
    RawDataBlockSize = 0



# The naming conventions follow those used in github.com/AstroDigitalVideo/ADVlib
# While not exactly 'Pythonic' (i.e., not snake case), it was my judgement (Bob Anderson) that
# fewer errors would be introduced and would reduce the intellectual 'load' of comparing
# the Python code against the equivalent code in github.com/AstroDigitalVideo/ADVlib

# We use type hinting so that it is easy to see the intent as matching the C++/C# code

# This module holds the relevant static methods (i.e., involved in reading a file as opposed
# to writing a file) that are defined in the C# file PInvoke.cs via
# 'public static class Adv2DLLlibs' (the pythonic equivalent is
# a module with the methods defined at the top level, such as this)


import pathlib
import platform
from ctypes import CDLL, byref, c_char_p, c_int, c_uint, pointer, c_int8, c_int16, c_int32, c_int64, c_float
from struct import pack, unpack
from typing import Tuple

from Adv2.Adv import AdvFileInfo, AdvFrameInfo, StreamId, TagPairType, Adv2TagType
from Adv2.AdvError import S_OK, ResolveErrorMessage, AdvLibException


# This mask is used to remove the sign bits from the 32 bit ret_val when it is converted to a Python int
RET_VAL_MASK = 0xffffffff

# The following code/tests will run at import time (startup) and raise/throw an exception if we can cannot
# distinguish Windows 64bit/32bit from Mac 64bit/32bit from Linux 64bit/32bit or find libraries.  We
# allow such an event to be a fatal error.

DLL_folder = pathlib.Path(__file__).parent / 'Adv2DLLlibs'

if platform.system().lower().startswith('windows'):
    # We're running on a on a Windows system
    if platform.architecture()[0] == '64bit':
        # The following line generates a platform agnostic filepath (deals with / \ issues)
        file_path = str(DLL_folder / 'AdvLib.Core64.dll')
        advDLL = CDLL(file_path)
    elif platform.architecture()[0] == '32bit':
        file_path = str(DLL_folder / 'AdvLib.Core32.dll')
        advDLL = CDLL(file_path)
    else:
        raise ImportError("System is neither 64 bit nor 32 bit.")
elif platform.system().lower().startswith('darwin'):
    # We're running on MacOS
    if platform.architecture()[0] == '64bit':
        # The following line generates a platform agnostic filepath (deals with / \ issues)
        file_path = str(DLL_folder / 'libAdvCore.dylib')
        advDLL = CDLL(file_path)
    elif platform.architecture()[0] == '32bit':
        raise ImportError("No 32 bit library available for MacOS")
    else:
        raise ImportError("System is neither 64 bit nor 32 bit.")
elif platform.system().lower().startswith('linux'):
    # We're running on a linux system
    if platform.architecture()[0] == '64bit':
        # The following line generates a platform agnostic filepath (deals with / \ issues)
        file_path = str(DLL_folder / 'libAdvCore64.so')
        advDLL = CDLL(file_path)
    elif platform.architecture()[0] == '32bit':
        # The following line generates a platform agnostic filepath (deals with / \ issues)
        file_path = str(DLL_folder / 'libAdvCore32.so')

        advDLL = CDLL(file_path)
    else:
        raise ImportError("System is neither 64 bit nor 32 bit.")


def AdvOpenFile(filepath: str, fileinfo: AdvFileInfo) -> int:
    advFileInfoFormat = '6iQiQi4BQ?4i'  # format of AdvFileInfo structure needed by pack() and unpack()
    # The above format specifies 6*int32 int64 int32 int64 4*uint8 int64 bool 4*uint32

    # Create a C style structure per the supplied format string
    file_info = pack(advFileInfoFormat,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

    # Convert the Python string to an array of bytes (with null terminator)
    file_path_as_bytes = bytes(filepath + '\0', 'utf-8')
    fname_ptr = c_char_p(file_path_as_bytes)  # fname_ptr equivalent to: const char* fileName

    # Call into the library with C-compatible arguments
    ret_val = advDLL.AdvOpenFile(fname_ptr, c_char_p(file_info)) & RET_VAL_MASK

    # Extract the entries from the the C struct file_info into the Python structure fileInfo
    fileinfo.Width = unpack(advFileInfoFormat, file_info)[0]
    fileinfo.Height = unpack(advFileInfoFormat, file_info)[1]
    fileinfo.CountMainFrames = unpack(advFileInfoFormat, file_info)[2]
    fileinfo.CountCalibrationFrames = unpack(advFileInfoFormat, file_info)[3]
    fileinfo.DataBpp = unpack(advFileInfoFormat, file_info)[4]
    fileinfo.MaxPixelValue = unpack(advFileInfoFormat, file_info)[5]
    fileinfo.MainClockFrequency = unpack(advFileInfoFormat, file_info)[6]
    fileinfo.MainStreamAccuracy = unpack(advFileInfoFormat, file_info)[7]
    fileinfo.CalibrationClockFrequency = unpack(advFileInfoFormat, file_info)[8]
    fileinfo.CalibrationStreamAccuracy = unpack(advFileInfoFormat, file_info)[9]
    fileinfo.MainStreamTagsCount = unpack(advFileInfoFormat, file_info)[10]
    fileinfo.CalibrationStreamTagsCount = unpack(advFileInfoFormat, file_info)[11]
    fileinfo.SystemMetadataTagsCount = unpack(advFileInfoFormat, file_info)[12]
    fileinfo.UserMetadataTagsCount = unpack(advFileInfoFormat, file_info)[13]
    fileinfo.UtcTimestampAccuracyInNanoseconds = unpack(advFileInfoFormat, file_info)[14]
    fileinfo.IsColourImage = unpack(advFileInfoFormat, file_info)[15]
    fileinfo.ImageLayoutsCount = unpack(advFileInfoFormat, file_info)[16]
    fileinfo.StatusTagsCount = unpack(advFileInfoFormat, file_info)[17]
    fileinfo.ImageSectionTagsCount = unpack(advFileInfoFormat, file_info)[18]
    fileinfo.ErrorStatusTagId = unpack(advFileInfoFormat, file_info)[19]

    return ret_val


def AdvCloseFile() -> int:
    return advDLL.AdvCloseFile() & RET_VAL_MASK


def AdvGetFileVersion(filepath: str) -> Tuple[str, int]:
    if not pathlib.Path(filepath).is_file():
        return f'Error - cannot find file: {filepath}', 0
    else:
        # Convert the Python string to an array of bytes (with null terminator)
        file_path_as_bytes = bytes(filepath + '\0', 'utf-8')
        fname_ptr = c_char_p(file_path_as_bytes)  # fname_ptr: const char* fileName

        version_num = advDLL.AdvGetFileVersion(fname_ptr) & RET_VAL_MASK

        if version_num == 0:
            return f'Error - not an FSTF file: {filepath}', 0
        else:
            return '', version_num


def AdvVer2_GetFramePixels(streamId: StreamId, frameNo: int,
                           pixels: any, frameInfo: AdvFrameInfo, systemErrorLen: int) -> int:
    advFrameInfoFormat = '7I4f3Bb8I'  # format of AdvFrameInfo needed by pack() and unpack()
    # The above format specifies 7 * uint32 4 * float32 3 * uchar char 8 * uint32
    # Create a C style structure per the supplied format string
    frame_info = pack(advFrameInfoFormat, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

    # Call into the library will C-compatible arguments
    ret_val = advDLL.AdvVer2_GetFramePixels(c_int(streamId.value), c_int(frameNo), pixels,
                                            c_char_p(frame_info), byref(c_int(systemErrorLen))) & RET_VAL_MASK

    # Unpack the C struct frame_info into the Python struct frameInfo
    frameInfo.StartTicksLo = unpack(advFrameInfoFormat, frame_info)[0]
    frameInfo.StartTicksHi = unpack(advFrameInfoFormat, frame_info)[1]
    frameInfo.EndTicksLo = unpack(advFrameInfoFormat, frame_info)[2]
    frameInfo.EndTicksHi = unpack(advFrameInfoFormat, frame_info)[3]

    frameInfo.UtcMidExposureTimestampLo = unpack(advFrameInfoFormat, frame_info)[4]
    frameInfo.UtcMidExposureTimestampHi = unpack(advFrameInfoFormat, frame_info)[5]
    frameInfo.Exposure = unpack(advFrameInfoFormat, frame_info)[6]

    frameInfo.Gamma = unpack(advFrameInfoFormat, frame_info)[7]
    frameInfo.Gain = unpack(advFrameInfoFormat, frame_info)[8]
    frameInfo.Shutter = unpack(advFrameInfoFormat, frame_info)[9]
    frameInfo.Offset = unpack(advFrameInfoFormat, frame_info)[10]

    frameInfo.GPSTrackedSatellites = unpack(advFrameInfoFormat, frame_info)[11]
    frameInfo.GPSAlmanacStatus = unpack(advFrameInfoFormat, frame_info)[12]
    frameInfo.GPSFixStatus = unpack(advFrameInfoFormat, frame_info)[13]
    frameInfo.GPSAlmanacOffset = unpack(advFrameInfoFormat, frame_info)[14]

    frameInfo.VideoCameraFrameIdLo = unpack(advFrameInfoFormat, frame_info)[15]
    frameInfo.VideoCameraFrameIdHi = unpack(advFrameInfoFormat, frame_info)[16]
    frameInfo.HardwareTimerFrameIdLo = unpack(advFrameInfoFormat, frame_info)[17]
    frameInfo.HardwareTimerFrameIdHi = unpack(advFrameInfoFormat, frame_info)[18]

    frameInfo.SystemTimestampLo = unpack(advFrameInfoFormat, frame_info)[19]
    frameInfo.SystemTimestampHi = unpack(advFrameInfoFormat, frame_info)[20]

    frameInfo.ImageLayoutId = unpack(advFrameInfoFormat, frame_info)[21]
    frameInfo.RawDataBlockSize = unpack(advFrameInfoFormat, frame_info)[22]

    return ret_val


def AdvVer2_GetTagPairValues(tagPairType: TagPairType, tagId: int) -> Tuple[int, str, str]:
    # Create big buffers of bytes to hold char string returns
    tagName = bytes('\0' * 256, 'utf8')
    tagValue = bytes('\0' * 256, 'utf8')
    ret_val = advDLL.AdvVer2_GetTagPairValues(
        c_int(tagPairType.value),
        c_int(tagId),
        c_char_p(tagName),
        c_char_p(tagValue)
    ) & RET_VAL_MASK
    tag_name_as_python_string = tagName.decode().strip('\0')  # This removes the null bytes
    tag_value_as_python_string = tagValue.decode().strip('\0')  # This removes the null bytes

    return ret_val, tag_name_as_python_string, tag_value_as_python_string


def AdvVer2_GetIndexEntries(mainIndex, calibrationIndex) -> int:
    return advDLL.AdvVer2_GetIndexEntries(mainIndex, calibrationIndex) & RET_VAL_MASK


def AdvVer2_GetStatusTagInfo(tagId: int) -> Tuple[Adv2TagType, str]:
    pTagNameSize = pointer(c_int(0))
    ret_val = advDLL.AdvVer2_GetStatusTagNameSize(c_uint(tagId), pTagNameSize) & RET_VAL_MASK

    if ret_val is not S_OK:
        raise AdvLibException(f'{ResolveErrorMessage(ret_val)}')

    tagNameSize = pTagNameSize.contents.value
    tagName = bytes('\0' * (tagNameSize + 1), 'utf8')
    pTagType = pointer(c_int(0))
    ret_val = advDLL.AdvVer2_GetStatusTagInfo(
        c_uint(tagId),
        c_char_p(tagName),
        pTagType
    ) & RET_VAL_MASK

    if ret_val is not S_OK:
        raise AdvLibException(f'{ResolveErrorMessage(ret_val)}')

    tagType = pTagType.contents.value
    return Adv2TagType(tagType), tagName.decode().strip('\0')


def AdvVer2_GetStatusTagUInt8(tagId: int) -> int:
    pTagValue = pointer(c_int8(0))
    ret_val = advDLL.AdvVer2_GetStatusTagUInt8(c_uint(tagId), pTagValue) & RET_VAL_MASK
    if ret_val is not S_OK:
        return -1
    return pTagValue.contents.value


def AdvVer2_GetStatusTagInt16(tagId: int) -> int:
    pTagValue = pointer(c_int16(0))
    ret_val = advDLL.AdvVer2_GetStatusTag16(c_uint(tagId), pTagValue) & RET_VAL_MASK
    if ret_val is not S_OK:
        return -1
    return pTagValue.contents.value


def AdvVer2_GetStatusTagInt32(tagId: int) -> int:
    pTagValue = pointer(c_int32(0))
    ret_val = advDLL.AdvVer2_GetStatusTag32(c_uint(tagId), pTagValue) & RET_VAL_MASK
    if ret_val is not S_OK:
        return -1
    return pTagValue.contents.value


def AdvVer2_GetStatusTagInt64(tagId: int) -> int:
    pTagValue = pointer(c_int64(0))
    ret_val = advDLL.AdvVer2_GetStatusTag64(c_uint(tagId), pTagValue) & RET_VAL_MASK
    if ret_val is not S_OK:
        return -1
    return pTagValue.contents.value


def AdvVer2_GetStatusTagReal(tagId: int) -> float:
    pTagValue = pointer(c_float(0))
    ret_val = advDLL.AdvVer2_GetStatusTagReal(c_uint(tagId), pTagValue) & RET_VAL_MASK
    if ret_val is not S_OK:
        return -1
    return pTagValue.contents.value


def AdvVer2_GetStatusTagUTF8String(tagId: int) -> str:
    pTagLen = pointer(c_int(0))
    ret_val = advDLL.AdvVer2_GetStatusTagSizeUTF8String(c_uint(tagId), pTagLen) & RET_VAL_MASK

    if ret_val is not S_OK:
        raise AdvLibException(f'{ResolveErrorMessage(ret_val)}')

    tagValue = bytes('\0' * (pTagLen.contents.value + 1), 'utf8')
    ret_val = advDLL.AdvVer2_GetStatusTagUTF8String(c_uint(tagId), c_char_p(tagValue)) & RET_VAL_MASK

    if ret_val is not S_OK:
        raise AdvLibException(f'{ResolveErrorMessage(ret_val)}')

    return tagValue.decode().strip('\0')


def GetLibraryVersion() -> str:
    libVer = bytes('\0' * 256, 'utf8')
    advDLL.GetLibraryVersion(c_char_p(libVer))
    return libVer.decode().strip('\0')


def GetLibraryPlatformId() -> str:
    platform_str = bytes('\0' * 256, 'utf8')
    advDLL.GetLibraryPlatformId(c_char_p(platform_str))
    return platform_str.decode().strip('\0')


def GetLibraryBitness() -> int:
    ret_val = advDLL.GetLibraryBitness()
    return ret_val



# The naming conventions follow those used in github.com/AstroDigitalVideo/ADVlib
# While not exactly 'Pythonic' (i.e., not snake case), it was my judgement (Bob Anderson) that
# fewer errors would be introduced and would reduce the intellectual 'load' of comparing
# the Python code against the equivalent code in github.com/AstroDigitalVideo/ADVlib


# Define a Exception that we can throw that automatically identifies itself
# as coming from the Adv2reader package.
class AdvLibException(Exception):
    pass


S_OK = 0x00000000


def ResolveErrorMessage(AdvResult: int, kind: str = 'human') -> str:
    # The dictionary provides strings that match the enum name of the error,
    # but also a more human readable form.
    if AdvResult in error_dict.keys():
        if kind == 'human':
            return error_dict[AdvResult][1]
        else:
            return error_dict[AdvResult][0]
    else:
        return f'0x{AdvResult:08X} is not a defined AdvResult'


error_dict = {
    0x81000001: ['E_ADV_NOFILE',
                 'The file could not be found.'],
    0x81000002: ['E_ADV_IO_ERROR',
                 'An I/O error has occurred.'],
    0x81001001: ['E_ADV_STATUS_ENTRY_ALREADY_ADDED',
                 'This status TagId has been already added to the current frame.'],
    0x81001002: ['E_ADV_INVALID_STATUS_TAG_ID',
                 'Unknown status TagId'],
    0x81001003: ['E_ADV_INVALID_STATUS_TAG_TYPE',
                 'The type of the status TagId doesn''t match the currently called method.'],
    0x81001004: ['E_ADV_STATUS_TAG_NOT_FOUND_IN_FRAME',
                 'The requested status TagId is not present in the current frame.'],
    0x81001005: ['E_ADV_FRAME_STATUS_NOT_LOADED',
                 'No status has been loaded. Call GetFramePixels() first.'],
    0x81001006: ['E_ADV_FRAME_NOT_STARTED',
                 'Frame not started. Call BeginFrame() first.'],
    0x81001007: ['E_ADV_IMAGE_NOT_ADDED_TO_FRAME',
                 'No image has been added to the started frame.'],
    0x81001008: ['E_ADV_INVALID_STREAM_ID',
                 'Invalid StreamId. Must be 0 for MAIN or 1 for CALIBRATION.'],
    0x81001009: ['E_ADV_IMAGE_SECTION_UNDEFINED',
                 'No Image Section has been defined.'],
    0x8100100A: ['E_ADV_STATUS_SECTION_UNDEFINED',
                 'No Status Section has been defined.'],
    0x8100100B: ['E_ADV_IMAGE_LAYOUTS_UNDEFINED',
                 'Image layouts undefined'],
    0x8100100C: ['E_ADV_INVALID_IMAGE_LAYOUT_ID',
                 'Invalid image LayoutId.'],
    0x8100100D: ['E_ADV_CHANGE_NOT_ALLOWED_RIGHT_NOW',
                 'This change is not allowed on an existing file or once a frame insertion has started.'],
    0x8100100E: ['E_ADV_IMAGE_SECTION_ALREADY_DEFINED',
                 'The Image Section can be only defined once per file'],
    0x8100100F: ['E_ADV_STATUS_SECTION_ALREADY_DEFINED',
                 'The Status Section can be only defined once per file'],
    0x81001010: ['E_ADV_IMAGE_LAYOUT_ALREADY_DEFINED',
                 'An Image Layout with this LayoutId has been already defined.'],
    0x81001011: ['E_ADV_INVALID_IMAGE_LAYOUT_TYPE',
                 'Invalid Image Layout type.'
                 ' Accepted values are FULL-IMAGE-RAW, 12BIT-IMAGE-PACKED and 8BIT-COLOR-IMAGE.'],
    0x81001012: ['E_ADV_INVALID_IMAGE_LAYOUT_COMPRESSION',
                 'Invalid Image Layout compression. Accepted values are UNCOMPRESSED, LAGARITH16 and QUICKLZ.'],
    0x81001013: ['E_ADV_INVALID_IMAGE_LAYOUT_BPP',
                 'Invalid Image Layout bits per pixel value. Accepted range is from 1 to 32.'],
    0x81001014: ['E_ADV_FRAME_MISSING_FROM_INDEX',
                 'The requested frame cannot be located in the index.'
                 ' The file may be have been corrupted. Try rebuilding it.'],
    0x81001015: ['E_ADV_FRAME_CORRUPTED',
                 'The frame binary data appears to be corrupted.'],
    0x81001016: ['E_ADV_FILE_NOT_OPEN',
                 'File system file is not open.'],
    0x81002001: ['E_ADV_NOT_AN_ADV_FILE',
                 'This is not a valid ADV file.'],
    0x81002002: ['E_ADV_VERSION_NOT_SUPPORTED',
                 'ADV version not supported.'],
    0x81002003: ['E_ADV_NO_MAIN_STREAM',
                 "'MAIN' stream is missing."],
    0x81002004: ['E_ADV_NO_CALIBRATION_STREAM',
                 "'CALIBRATION' stream is missing."],
    0x81002005: ['E_ADV_TWO_SECTIONS_EXPECTED',
                 'ADV files must have two sections.'],
    0x81002006: ['E_ADV_NO_IMAGE_SECTION',
                 "First section must be 'IMAGE'."],
    0x81002007: ['E_ADV_NO_STATUS_SECTION',
                 "Second section must be 'STATUS'."],
    0x81002008: ['E_ADV_IMAGE_SECTION_VERSION_NOT_SUPPORTED',
                 "'IMAGE' section version is not supported."],
    0x81002009: ['E_ADV_IMAGE_LAYOUT_VERSION_NOT_SUPPORTED',
                 'Image Layout version is not supported.'],
    0x8100200A: ['E_ADV_STATUS_SECTION_VERSION_NOT_SUPPORTED',
                 "'STATUS' section version is not supported."],
    0x71000001: ['S_ADV_TAG_REPLACED',
                 'An existing tag with the same name has been replaced.'],
    0x80004005: ['E_FAIL',
                 'Error.'],
    0x80004001: ['E_NOTIMPL',
                 'Not Implemented'],
    0x00000000: ['S_OK',
                 'Success.'],
}



from setuptools import setup
import pathlib
from Adv2 import Version


VERSION = Version.version()

KEYWORDS = ["ADV file reader", 'Astro Digital Video version 2 file reader']

CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Science/Research",
    "Natural Language :: English",
    "License :: OSI Approved :: MIT License",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering",
]

INSTALL_REQUIRES = ['numpy']

HERE = pathlib.Path.cwd()

README = (HERE / "README.md").read_text()

setup(
    name="Adv2",
    python_requires='>=3.7',
    description='Adv2reader reads version 2 Astro Digital Video files.',
    license='License :: OSI Approved :: MIT License',
    url=r'https://github.com/bob-anderson-ok/adv2reader',
    version=VERSION,
    author='Bob Anderson',
    author_email='bob.anderson.ok@gmail.com',
    maintainer='Bob Anderson',
    maintainer_email='bob.anderson.ok@gmail.com',
    keywords=KEYWORDS,
    long_description=README,
    long_description_content_type="text/markdown",
    packages=["Adv2"],
    include_package_data=True,
    classifiers=CLASSIFIERS,
    install_requires=INSTALL_REQUIRES,
)
